/* df3vt.c
 *
 * DF3 files as SQLite virtual tables.
 *
 * based on the SQLite 'ext/mic/csv.c' loadable extension,
 * the orginal copyright notice reads:
 *
 * 2016-05-28
 *
 * The author disclaims copyright to this source code.  In place of
 * a legal notice, here is a blessing:
 *
 *    May you do good and not evil.
 *    May you find forgiveness for yourself and forgive others.
 *    May you share freely, never taking more than you give.
 *
 * -------------------------------------------------------------------------- *
 *
 * version 0.1.0
 *
 * building the loadable extension (assuming GCC):
 *
 *     $ cc -std=c99 -pedantic -Werror -O2 -fPIC -o df3vt.so \
 *               df3vt.c -shared -ldf3
 *
 * usage, from the SQLite shell:
 *
 *    > .load /path/to/df3vt
 *    > CREATE VIRTUAL TABLE temp.df3 USING df3vt(file=FILENAME, readonly=yes);
 *    > SELECT * FROM temp.df3;
 *
 * the arguments to the parameters are:
 *    file     - a POSIX.1 compliant file name, '.df3' extension required.
 *               mandatory.
 *    readonly - a boolean value.  optional parameter, default true.
 *
 * the columns are named 'x', 'y', 'z', and 'v'.
 *
 * the df3 table does not support INSERTs or DELETEs.
 * UPDATEs will be added in a later release, but will, necessarily,
 * be limited to the 'v' column.
 * ORDER BY processing too will be added in a later release.
 */

#include <df3.h>
#include <errno.h>
#include <regex.h>
#include <sqlite3ext.h>
SQLITE_EXTENSION_INIT1
#include <string.h>

#ifndef SQLITE_OMIT_VIRTUALTABLE

/* self */
#define TOKEN  "df3vt"

#define AUU  __attribute__((unused))

/* known-to-work minimum version */
#define MIN_LIB_VERSION  3008007

/* no column constraints (ignored on registration)  */
#define VTSCHEMA  "CREATE TABLE x (x INT, y INT, z INT, v INT);"

/* use "keyword = value" style parameters */
#define REEXPR   "^[[:space:]]+?([[:alpha:]]+)[[:space:]]+?=" \
                 "[[:space:]]+?['\"]?([[:alpha:][:digit:]_.,/-]+)['\"]?$"
#define REFLAGS  (REG_EXTENDED|REG_ICASE)

static regex_t kvre;

enum df3ArgKey {
  kw_badfmt = -1,
  kw_unknown = 0,
  kw_file,
  kw_readonly,
  kw_unused
};
typedef enum df3ArgKey df3ArgKey;

/* need distinct initials, cf reParse() */
static char const * const kwords[] = { "", "file", "readonly", NULL };

/* an instance of a virtual table */
typedef struct df3Vtable df3Vtable;
struct df3Vtable {
  sqlite3_vtab base;    /* base class */
  df3Handle df3;        /* handle */
  bool ro;              /* readonly arg */
  bool dirty;           /* whether updated */
};

/* cursor for a virtual table */
typedef struct df3Vcursor df3Vcursor;
struct df3Vcursor {
  sqlite3_vtab_cursor base;    /* base class */
  df3Coord at;                 /* current pos */
  uint32_t atval;              /* current pos voxel */
  df3Handle *df3;              /* pointer to handle */
  sqlite3_int64 rowid;         /* current row.  negative == EOF */
};

/* --[argument processing]-------------------------------------------------- */

static bool checkBoolVal(char const * const arg, bool *rf) {
  char const * const f[] = { "false", "no", "disable", "off", "0", NULL };
  for (size_t i = 0; f[i]; ++i)
    if (!strcasecmp(arg, f[i])) { *rf = false; return true; }
  char const * const t[] = { "true", "yes", "enable", "on", "1", NULL };
  for (size_t i = 0; t[i]; ++i)
    if (!strcasecmp(arg, t[i])) { *rf = true; return true; }
  return false;
}

static df3ArgKey reParse(regex_t *re,
                         char const * const arg,
                         size_t *so,
                         size_t *eo) {
  regmatch_t m[3] = {{0,0}};  /* arg + 2 subs */
  if (regexec(re, arg, 3, m, 0))
    return kw_badfmt;
  char const * const p = arg + m[1].rm_so;
  df3ArgKey k = kw_unknown;
  for (size_t i = 0; kwords[i]; ++i)
    if (!strncasecmp(p, kwords[i], (size_t)(m[1].rm_eo - m[1].rm_so))) {
      k = (df3ArgKey)i;
      break;
    }
  if (kw_unknown != k) {
    *so = (size_t)m[2].rm_so;
    *eo = (size_t)m[2].rm_eo;
  }
  return k;
}

static bool procParams(regex_t * const re,
                       char const * const args[],
                       df3Vtable *tbl,
                       char **emsg) {
  if (!args[0]) {
    *emsg = sqlite3_mprintf("error missing 'file' parameter.");
    return false;
  }
  size_t so, eo;
  so = eo = 0;
  df3ArgKey t = reParse(re, args[0], &so, &eo);
  if (kw_file != t) {
    *emsg = sqlite3_mprintf("error first arg must be 'file'.");
    return false;
  }
  else {
    size_t const n = eo - so;
    char *tmp = sqlite3_malloc64(n + 1);
    if (!tmp) return false;
    memset(tmp, 0, n + 1);
    strncpy(tmp, args[0] + so, n);
    if (!df3_check_fname(tmp)) {
      *emsg = sqlite3_mprintf("error (unusable) file '%s'.", tmp);
      sqlite3_free(tmp);
      return false;
    }
    tbl->df3.fname = tmp;
  }

  if (args[1]) {
    bool f = true;
    so = eo = 0;
    t = reParse(re, args[1], &so, &eo);
    if (kw_readonly != t) {
      *emsg = sqlite3_mprintf("error stray parameter.");
      return false;
    }
    char buf[1024] = {'\0'};
    strncpy(buf, args[1] + so, eo - so);
    if (!checkBoolVal(buf, &f)) {
      *emsg = sqlite3_mprintf("error 'readonly' needs boolean arg.");
      return false;
    }
    tbl->ro = f;
  }

  return true;
}

/* --[module interface functions]------------------------------------------- */

/* df3vtDisconnect()
 * -------------------------------------------------------------------------- *
 * the xDisconnect and xDestroy methods are the same since there is no
 * backing store to create.
 */
static int df3vtDisconnect(sqlite3_vtab *tbl) {
  df3Vtable *p = (df3Vtable*)tbl;
  (void)df3_close_file(&p->df3);
  sqlite3_free(p->df3.fname);
  sqlite3_free(p);
  return SQLITE_OK;
}

/* df3vtConnect()
 * -------------------------------------------------------------------------- *
 * the xConnect and xCreate methods are the same since there is no backing
 * store to create.
 */
static int df3vtConnect(sqlite3 *db,
                        void *clientdata,
                        int argc AUU,
                        const char * const *argv,
                        sqlite3_vtab **vtbl,
                        char **emsg) {
  df3Vtable *p = sqlite3_malloc64(sizeof(df3Vtable));
  if (!p) return SQLITE_ERROR;
  memset(p, 0, sizeof(df3Vtable));

  p->ro = true;  /* default */
  regex_t * const re = (regex_t *)clientdata;
  char *temsg = NULL;
  if (!procParams(re, &argv[3], p, &temsg)) {
    if (*emsg) sqlite3_free(*emsg);
    *emsg = temsg;
    goto label_fail;
  }

  df3Error const ec = df3_open_file(&p->df3, p->ro);
  if (ecOk != ec) {
    if (*emsg) sqlite3_free(*emsg);
    *emsg = sqlite3_mprintf("error opening '%s': %s",
            p->df3.fname, df3_error_message(ec));
    goto label_fail;
  }

  *vtbl = (sqlite3_vtab *)p;
  if (sqlite3_declare_vtab(db, VTSCHEMA))
    goto label_fail;

  return SQLITE_OK;

label_fail:
  if (p) df3vtDisconnect((sqlite3_vtab *)p);
  return SQLITE_ERROR;
}

/* df3vtClose()
 * -------------------------------------------------------------------------- *
 * destructor for df3Vcursor.
 */
static int df3vtClose(sqlite3_vtab_cursor *curs) {
  sqlite3_free(curs);
  return SQLITE_OK;
}

/* df3vtOpen()
 * -------------------------------------------------------------------------- *
 * constructor for new df3Vcursor.  nb pointer to parent "table".
 */
static int df3vtOpen(sqlite3_vtab *vtbl, sqlite3_vtab_cursor **pcurs) {
  df3Vtable *p = (df3Vtable*)vtbl;
  df3Vcursor *tmp = sqlite3_malloc64(sizeof(df3Vcursor));
  if (!tmp) return SQLITE_NOMEM;
  memset(tmp, 0, sizeof(df3Vcursor));
  tmp->df3 = &p->df3;
  tmp->rowid = -1;
  *pcurs = (sqlite3_vtab_cursor *)tmp;
  return SQLITE_OK;
}

/* df3vtColumn()
 * -------------------------------------------------------------------------- *
 * return value of given column for the current row.
 */
static int df3vtColumn(sqlite3_vtab_cursor *curs, sqlite3_context *ctx, int n) {
  df3Vcursor *p = (df3Vcursor*)curs;
  switch (n) {
    case 0:  sqlite3_result_int(ctx, (int)p->at.x); break;
    case 1:  sqlite3_result_int(ctx, (int)p->at.y); break;
    case 2:  sqlite3_result_int(ctx, (int)p->at.z); break;
    case 3:  sqlite3_result_int64(ctx, p->atval); break;
    default: return SQLITE_CONSTRAINT_VTAB;
  }
  return SQLITE_OK;
}

/* df3vtRowid()
 * -------------------------------------------------------------------------- *
 * return the rowid for the current row.  no eof check.
 */
static int df3vtRowid(sqlite3_vtab_cursor *curs, sqlite_int64 *rowid) {
  df3Vcursor *p = (df3Vcursor*)curs;
  *rowid = p->rowid;
  return SQLITE_OK;
}

/* df3vtEof()
 * -------------------------------------------------------------------------- *
 * true if the cursor has been moved past last voxel.
 */
static int df3vtEof(sqlite3_vtab_cursor *curs) {
  df3Vcursor *p = (df3Vcursor*)curs;
  return 0 > p->rowid;
}

/* df3vtNext()
 * -------------------------------------------------------------------------- *
 * advance cursor to its next row of input.
 * set eof marker when the end of input.
 */
static sqlite3_int64 at2rowid(df3Axes const axes, df3Coord const at) {
  /* nb ignore bpv */
  size_t const r = at.z * axes.x * axes.y + at.y * axes.x + at.x;
  return (sqlite3_int64)r;
}

static int df3vtNext(sqlite3_vtab_cursor *curs) {
  df3Vcursor *p = (df3Vcursor*)curs;

  if (-1 != p->rowid) {
    bool atend = false;
    p->at.x += 1;
    if (p->df3->axes.x == p->at.x) {
      p->at.x = 0;
      p->at.y += 1;
      if (p->df3->axes.y == p->at.y) {
        p->at.y = 0;
        p->at.z += 1;
        if (p->df3->axes.z == p->at.z) {
          p->at.z = 0;
          atend = true;
        }
      }
    }
    if (atend)
      p->rowid = -1;
    else {
      uint32_t v = 0;
      df3Error const ec = df3_get_voxel(p->df3, p->at, &v);
      if (ecOk != ec) return SQLITE_IOERR_SHORT_READ;
      p->atval = v;
      p->rowid = at2rowid(p->df3->axes, p->at);
    }
  }

  return SQLITE_OK;
}

/* df3vtFilter()
 * -------------------------------------------------------------------------- *
 * only a full table scan is supported.  xFilter simply goes to 0,0,0.
 */
static int df3vtFilter(sqlite3_vtab_cursor *curs, 
                       int idxNum AUU,
                       const char *idxStr AUU,
                       int argc AUU,
                       sqlite3_value **argv AUU) {
  df3Vcursor *p = (df3Vcursor*)curs;
  p->at = (df3Coord){0,0,0};
  uint32_t v = 0;
  df3Error const ec = df3_get_voxel(p->df3, p->at, &v);
  if (ecOk != ec) return SQLITE_IOERR_SHORT_READ;
  p->atval = v;
  p->rowid = 0;
  return SQLITE_OK;
}

/* df3vtBestIndex()
 * -------------------------------------------------------------------------- *
 * only a forward full table scan is supported.
 * xBestIndex curently a no-op.
*/
static int df3vtBestIndex(sqlite3_vtab *tbl AUU, sqlite3_index_info *inf AUU) {
  return SQLITE_OK;
}

/* df3vtUpdate()
 * -------------------------------------------------------------------------- *
 */
static int df3vtUpdate(sqlite3_vtab *tbl AUU,
                       int todo1 AUU,
                       sqlite3_value **todo2 AUU,
                       sqlite3_int64 *todo3 AUU) {
  return SQLITE_OK;
}

/* mandatory interface */
static sqlite3_module df3vtModule = {
  0,                /* iVersion */
  df3vtConnect,     /* xCreate */
  df3vtConnect,     /* xConnect */
  df3vtBestIndex,   /* xBestIndex */
  df3vtDisconnect,  /* xDisconnect */
  df3vtDisconnect,  /* xDestroy */
  df3vtOpen,        /* xOpen */
  df3vtClose,       /* xClose */
  df3vtFilter,      /* xFilter */
  df3vtNext,        /* xNext */
  df3vtEof,         /* xEof */
  df3vtColumn,      /* xColumn */
  df3vtRowid,       /* xRowid */
  df3vtUpdate,      /* xUpdate */
  0,                /* xBegin */
  0,                /* xSync */
  0,                /* xCommit */
  0,                /* xRollback */
  0,                /* xFindMethod */
  0,                /* xRename */
  0,                /* xSavepoint */
  0,                /* xRelease */
  0                 /* xRollbackTo */
};

#endif /* !defined(SQLITE_OMIT_VIRTUALTABLE) */

/* -------------------------------------------------------------------------- *
 * this routine is called when the extension is loaded.  the new df3vt virtual
 * table module is registered with the calling database connection.
 * currently only have one "client data" item (RE), so no wrapping struct.
 */
static void tidyDf3vt(void *clientdata AUU) {
  regfree(&kvre);
  return;
}

/* when built with sqlite3_df3vt_init, sqlite3 complains about not
 * finding "sqlite3_dfvt_init"!!  why?
 */
#ifdef _WIN32
__declspec(dllexport)
#endif
int sqlite3_extension_init(sqlite3 *db, 
                           char **emsg,
                           const sqlite3_api_routines *api) {
#ifndef SQLITE_OMIT_VIRTUALTABLE	
  if (MIN_LIB_VERSION > SQLITE_VERSION_NUMBER)
    return SQLITE_FORMAT;
  else if (regcomp(&kvre, REEXPR, REFLAGS)) {
    int const e = errno;
    if (*emsg) sqlite3_free(*emsg);
    *emsg = sqlite3_mprintf("error creating RE: %s.", strerror(e));
    return SQLITE_ERROR;
  }
  SQLITE_EXTENSION_INIT2(api);
  return sqlite3_create_module_v2(db, TOKEN, &df3vtModule, &kvre, tidyDf3vt);
#else
  return SQLITE_OK;
#endif
}

/* ----------------------------------------------------------------- *
 *  this file is part of the 'df3 tools' project.                    *
 *  the content above is covered by the GPLv3+, see file COPYING.    *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.              *
 *  all rights reserved.                                             *
 * ----------------------------------------------------------------- */
